@@ -169,7 +169,7 @@ func runBundleVetCmd(cmd *cobra.Command, args []string) error {
 			return err
 		}
 
-		log = LoggerBundle(logr.NewContext(cmd.Context(), log), bundle.Name, apiv1.RuntimeDefaultName)
+		log = loggerBundle(logr.NewContext(cmd.Context(), log), bundle.Name, apiv1.RuntimeDefaultName, true)
 
 		if len(bundle.Instances) == 0 {
 			return fmt.Errorf("no instances found in bundle")
@@ -193,7 +193,7 @@ func runBundleVetCmd(cmd *cobra.Command, args []string) error {
 				if i.Namespace == "" {
 					return fmt.Errorf("instance %s does not have a namespace", i.Name)
 				}
-				log := LoggerBundleInstance(logr.NewContext(cmd.Context(), log), bundle.Name, cluster.Name, i.Name)
+				log := loggerBundleInstance(logr.NewContext(cmd.Context(), log), bundle.Name, cluster.Name, i.Name, true)
 				log.Info("instance is valid")
 			}
 		}
