@@ -28,6 +28,7 @@ import (
 
 	apiv1 "github.com/stefanprodan/timoni/api/v1alpha1"
 	"github.com/stefanprodan/timoni/internal/engine"
+	"github.com/stefanprodan/timoni/internal/logger"
 	"github.com/stefanprodan/timoni/internal/runtime"
 )
 
@@ -93,7 +94,7 @@ func runRuntimeBuildCmd(cmd *cobra.Command, args []string) error {
 	}
 
 	for _, cluster := range clusters {
-		log := LoggerRuntime(cmd.Context(), rt.Name, cluster.Name)
+		log := loggerRuntime(cmd.Context(), rt.Name, cluster.Name, true)
 
 		kubeconfigArgs.Context = &cluster.KubeContext
 		rm, err := runtime.NewResourceManager(kubeconfigArgs)
@@ -116,7 +117,7 @@ func runRuntimeBuildCmd(cmd *cobra.Command, args []string) error {
 		sort.Strings(keys)
 
 		for _, k := range keys {
-			log.Info(fmt.Sprintf("%s: %s", colorizeSubject(k), values[k]))
+			log.Info(fmt.Sprintf("%s: %s", logger.ColorizeSubject(k), values[k]))
 		}
 
 		if len(values) == 0 {
